from typing import Any, NamedTuple, Union

import jax.numpy as jnp

from jaxtyping import Array, ArrayLike

from .sparse import CenteredSparseMatrix, SparseMatrix


DataMatrix = Array | SparseMatrix | CenteredSparseMatrix
FloatOrArray = Union[float, ArrayLike]

class ELBOResults(NamedTuple):

    """Define the class of all components in ELBO.

    Attributes:
        elbo: the value of ELBO
        E_ll: Expectation of log-likelihood
        negKL_z: -KL divergence of Z
        negKL_w: -KL divergence of W
        negKL_gamma: -KL divergence of gamma

    """

    elbo: FloatOrArray
    E_ll: FloatOrArray
    negKL_z: FloatOrArray
    negKL_w: FloatOrArray
    negKL_gamma: FloatOrArray

    def __str__(self):
        return (
            f"ELBO = {self.elbo:.3f} | E[logl] = {self.E_ll:.3f} | "
            f"-KL[Z] = {self.negKL_z:.3f} | -KL[W] = {self.negKL_w:.3f} | "
            f"-KL[G] = {self.negKL_gamma:.3f}"
        )


class ModelParams(NamedTuple):
    """
    Define the class for variational parameters of all the variable we need
    to infer from the perturbVI.

    Attributes:
        mean_z: mean parameter for factor Z
        var_z: variance parameter for factor Z
        mean_w: conditional mean parameter for loadings W
        var_w: conditional variance parameter for loading W
        alpha: parameter for the gamma that follows multinomial
                distribution
        tau: inverse variance parameter of observed data X
        tau_0: inverse variance parameter of single effect w_kl
        theta: parameter for annotation prior
        pi: prior probability for gamma
        ann_state: internal state for learning theta
        mean_beta: mean parameters for perturbation effect matrix
        var_beta: variance parameters for perturbation effect matrix
        tau_beta: inverse variance parameters for perturbation effect matrix
        p: prior for Eta from spike-and-slab prior
        p_hat: variational parameters for Eta
    """

    # ssq
    x_ssq: Array

    # variational params for Z
    mean_z: Array
    var_z: Array

    # variational params for W given Gamma
    mean_w: Array
    var_w: Array

    @property
    def n_dim(self):
        return self.mean_z.shape[0]

    @property
    def z_dim(self):
        return self.mean_z.shape[1]

    @property
    def p_dim(self):
        return self.mean_w.shape[2]

    # variational params for Gamma
    alpha: Array

    # residual precision param
    tau: FloatOrArray
    tau_0: Array

    # prior probability for gamma
    theta: Array
    pi: Array
    # internal state for learning theta
    ann_state: Any

    # variational params of perturbation effects
    mean_beta: Array
    var_beta: Array

    # residual precision for beta
    tau_beta: Array

    # prior for Eta
    p: Array
    # variational params for Eta
    p_hat: Array

    @property
    def W(self) -> Array:
        return jnp.sum(self.mean_w * self.alpha, axis=0)



from datetime import datetime
from functools import partial

import pandas as pd
import numpy as np
import pickle
import os

import equinox as eqx
import jax.scipy.special as jspec
import lineax as lx

from jax import jit, lax, numpy as jnp, random as rdm
from jaxtyping import Array

multi_linear_solve = eqx.filter_vmap(lx.linear_solve, in_axes=(None, 1, None))

_add_ufunc = jnp.frompyfunc(jnp.add, nin=2, nout=1, identity=0)
outer_add = _add_ufunc.outer


def logdet(A: Array) -> Array:
    _, ldet = jnp.linalg.slogdet(A)
    return ldet


def kl_discrete(alpha: Array, pi: Array) -> Array:
    """A function that calculates the Kullback-Leibler divergence for multinomial distributions

    **Arguments:**

    -`alpha` [`Array`]: An array representing the first discrete distribution.

    -`pi` [`Array`]: An array representing the second discrete distribution.

    **Returns:**

    The Kullback-Leibler divergence between the two distributions.
    """
    return jnp.sum(jspec.xlogy(alpha, alpha) - jspec.xlogy(alpha, pi))


@partial(jit, static_argnums=(2, 3, 4))
def prob_pca(rng_key, X, k, max_iter=1000, tol=1e-3):
    """Probabilistic PCA algorithm to initialize latent factors.

    **Arguments:**

    -`rng_key` [`PRNGKey`]: Random key generator.

    -`X` [`Array`]: The observed data.

    -`k` [`int`]: The latent dimension.

    -`max_iter` [`int`]: The maximum number of iterations, default is 1000.

    -`tol` [`float`]: The convergence tolerance, default is 1e-3.

    **Returns:**

    - `Z` [`Array`]: The estimated latent factors.

    -`W` [`Array`]: The estimated loadings.
    """

    n_dim, p_dim = X.shape

    # initial guess for W
    w_key, z_key = rdm.split(rng_key, 2)

    # good enough for initialization
    solver = lx.Cholesky()

    # check if reach the max_iter, or met the norm criterion every 100 iteration
    def _condition(carry):
        i, _, Z, old_Z = carry
        iter_check = i < max_iter
        tol_check = jnp.linalg.norm(Z - old_Z) > tol
        # scaled_tol_check = tol_check / n_dim > tol
        return iter_check & tol_check

    # EM algorithm for PPCA
    def _step(carry):
        i, W, Z, _ = carry

        # E step
        W_op = lx.MatrixLinearOperator(W @ W.T, tags=lx.positive_semidefinite_tag)
        Z_new = multi_linear_solve(W_op, W @ X.T, solver).value

        # M step
        Z_op = lx.MatrixLinearOperator(Z_new.T @ Z_new, tags=lx.positive_semidefinite_tag)
        W = multi_linear_solve(Z_op, Z_new.T @ X, solver).value.T

        return i + 1, W, Z_new, Z

    W = rdm.normal(w_key, shape=(k, p_dim))
    Z = rdm.normal(z_key, shape=(n_dim, k))
    Z_zero = jnp.zeros_like(Z)
    initial_carry = 0, W, Z, Z_zero

    _, W, Z, _ = lax.while_loop(_condition, _step, initial_carry)
    Z, _ = jnp.linalg.qr(Z)

    return Z, W


# Create function to evaluate Local False Sign Rate
# First Create a function to sample single effect matrix based on params.alpha
def bern_sample(alpha):
    """Sample from a Bernoulli distribution with probability alpha.

    **Arguments:**

    - `alpha` [`Array`]: The probability of each row in the L x K matrix.

    **Returns:**

    - `efficient_result_matrix` [`Array`]: The sampled matrix.

    """
    l_dim, z_dim, _ = alpha.shape
    # Generate random numbers for each row in the L x K matrix
    # These random numbers are used as indices for selecting features
    random_indices = np.random.rand(l_dim, z_dim)
    # Calculate the cumulative sum of probabilities along the P dimension
    cumulative_probabilities = np.cumsum(alpha, axis=2)

    # Determine the indices in P where the cumulative probability exceeds the random index
    # This effectively samples from the probability distribution
    feature_indices = np.argmax(cumulative_probabilities > random_indices[..., np.newaxis], axis=2)
    # Initialize the result matrix with zeros
    efficient_result_matrix = np.zeros_like(alpha)

    # Use advanced indexing to set the selected features to 1
    efficient_result_matrix[np.arange(l_dim)[:, np.newaxis], np.arange(z_dim), feature_indices] = 1

    return efficient_result_matrix


def bern_sample_jax(key, alpha):
    """JAX version of bern_sample function.

    Arguments:
        key: JAX random key
        alpha: probability matrix of shape (l_dim, z_dim, p_dim)
    """
    random_values = rdm.uniform(key, shape=alpha.shape[:-1])
    cumsum = jnp.cumsum(alpha, axis=-1)
    return jnp.eye(alpha.shape[-1])[jnp.argmax(cumsum > random_values[..., None], axis=-1)]


@partial(jit, static_argnums=(2,))
def _compute_lfsr_step(key, params, iters):
    """Jitted inner loop of LFSR computation"""
    l_dim, z_dim, p_dim = params.alpha.shape
    g_dim, _ = params.mean_beta.shape
    reshaped_var_w = jnp.repeat(params.var_w[:, :, jnp.newaxis], p_dim, axis=2)

    def _inner_loop(carry, i):  # Modified to accept iteration index
        key, total_pos, total_neg = carry

        # Split keys for different random operations
        key, w_key, a_key, e_key, b_key = rdm.split(key, 5)

        # Sample W
        sample_w = params.mean_w + jnp.sqrt(reshaped_var_w) * rdm.normal(w_key, shape=params.mean_w.shape)
        sample_alpha = bern_sample_jax(a_key, params.alpha)
        sample_W = jnp.sum(sample_w * sample_alpha, axis=0)

        # Sample B
        sample_eta = rdm.bernoulli(e_key, params.p_hat.T)
        sample_beta = params.mean_beta + jnp.sqrt(params.var_beta) * rdm.normal(b_key, shape=params.mean_beta.shape)
        sample_B = sample_beta * sample_eta

        # Compute outer product
        sample_oe = sample_B @ sample_W
        ind_pos = sample_oe >= 0
        ind_neg = sample_oe <= 0

        return (key, total_pos + ind_pos, total_neg + ind_neg), None

    # Initialize
    total_pos_zero = jnp.zeros((g_dim, p_dim))
    total_neg_zero = jnp.zeros((g_dim, p_dim))
    init_carry = (key, total_pos_zero, total_neg_zero)

    # Run the loop
    (_, total_pos_zero, total_neg_zero), _ = lax.scan(_inner_loop, init_carry, jnp.arange(iters))

    return total_pos_zero, total_neg_zero


def compute_lfsr(key, params, iters=2000):
    """Compute the LFSR (Local False Sign Rate) using the given parameters.

    Arguments:
        key: JAX random key
        params: The parameters of the model
        iters: Number of iterations (default=2000)
    """
    current_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    print(f"Start computing LFSR at {current_time}")

    # Split computation into chunks to show progress
    chunk_size = 100
    num_chunks = iters // chunk_size
    remaining = iters % chunk_size

    total_pos = 0
    total_neg = 0

    for i in range(num_chunks):
        pos_chunk = 0
        neg_chunk = 0
        # Process each iteration within the chunk individually
        for j in range(chunk_size):
            iter_key = rdm.fold_in(key, i * chunk_size + j)  # Unique key for each iteration
            pos, neg = _compute_lfsr_step(iter_key, params, 1)  # Process single iteration
            pos_chunk += pos
            neg_chunk += neg
        total_pos += pos_chunk
        total_neg += neg_chunk
        print(f"Completed {(i + 1) * chunk_size}/{iters} iterations")

    # Handle remaining iterations if any
    if remaining > 0:
        pos_rem = 0
        neg_rem = 0
        for j in range(remaining):
            iter_key = rdm.fold_in(key, num_chunks * chunk_size + j)
            pos, neg = _compute_lfsr_step(iter_key, params, 1)
            pos_rem += pos
            neg_rem += neg
        total_pos += pos_rem
        total_neg += neg_rem
        print(f"Completed {iters}/{iters} iterations")

    # Compute final LFSR
    lfsr = jnp.minimum(total_pos, total_neg) / iters

    current_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    print(f"Finished computing LFSR at {current_time}")

    return lfsr


def pip_analysis(pip: jnp.ndarray, rho=0.9, rho_prime=0.05):
    """Create a function to give a quick summary of PIPs

    Args:
        pip:the pip matrix, a ndarray from results object returned by
        infer.perturbvi

    """
    z_dim, p_dim = pip.shape
    results = []

    print(f"Of {p_dim} features from the data, SuSiE PCA identifies:")
    for k in range(z_dim):
        num_signal = jnp.where(pip[k, :] >= rho)[0].shape[0]
        num_zero = jnp.where(pip[k, :] < rho_prime)[0].shape[0]
        print(f"Component {k} has {num_signal} features with pip>{rho}; and {num_zero} features with pip<{rho_prime}")
        results.append([num_signal, num_zero])

    df = pd.DataFrame(results, columns=["num_signal", "num_zero"])

    # Calculate and print mean and standard deviation for each column
    mean_signal = df["num_signal"].mean()
    std_signal = df["num_signal"].std()
    mean_zero = df["num_zero"].mean()
    std_zero = df["num_zero"].std()

    print(f"Mean and standard deviation for num_signal: {mean_signal}, {std_signal}")
    print(f"Mean and standard deviation for num_zero: {mean_zero}, {std_zero}")

    return df


def find_top_genes(df, pip_cutoff = 0.9):
    high_value_genes = {}
    for column in df.columns:
        high_values = df[df[column] > pip_cutoff].index.tolist()
        high_value_genes[column] = high_values
    return high_value_genes

def luhmes_analysis(dir: str, guide_path: str, gene_path: str):
    """Create a function to give a quick summary of LUHMES results
    Args:
        dir: Results directory
        guide_path: Guide file path (csv)
        gene_path: Gene symbol path (csv)
    """
    params_path = f"{dir}/params_file.pkl"
    pip_path = f"{dir}/pip.txt"
    W_path = f"{dir}/W.txt"
    pip_df_path = f"{dir}/pip.csv"
    lfsr_path = f"{dir}/lfsr.csv"
    p_hat_path = f"{dir}/p_hat.csv"
    beta_path = f"{dir}/beta_target.csv"
    overall_path = f"{dir}/overall_effect.csv"

    # guide_path = "luhmes_G.csv"
    # gene_path = "luhmes_gene_symbol_v2.csv"
    G = pd.read_csv(guide_path, index_col=0)
    G_reduce = G.drop(columns=["Nontargeting"])
    perturbed = G_reduce.columns.to_list()  # 14 genes perturbed
    genes = pd.read_csv(gene_path, header=None)[0].to_list()  # 6000 gene symbols (background genes)

    with open(params_path, "rb") as file:
        params = pickle.load(file)

    pip = np.loadtxt(pip_path)
    W = pd.DataFrame(np.loadtxt(W_path))

    z_dim, p_dim = params.W.shape
    g_dim, z_dim = params.mean_beta.shape
    n_dim, z_dim = params.mean_z.shape
    print(f"Dims: z_dim={z_dim}, p_dim={p_dim}, g_dim={g_dim}, n_dim={n_dim}\n")
    column_names_b = [f"b{i}" for i in range(z_dim)]
    column_names_w = [f"w{i}" for i in range(z_dim)]

    beta_sparse = params.mean_beta * params.p_hat.T
    overall_effect = beta_sparse @ params.W

    # pip_df = pd.DataFrame(pip.T, columns=column_names_w)
    pip_df = pd.DataFrame(pip.T, columns=column_names_w, index=genes)
    pip_df.to_csv(pip_df_path)
    
    analysis = pip_analysis(jnp.asarray(pip).astype(jnp.float32),rho=0.90,rho_prime=0.10)

    for col in pip_df.columns:
        sig_genes = pip_df.index[pip_df[col] > 0.9].tolist()
        if not os.path.exists(f"{dir}/pip"):
            os.makedirs(f"{dir}/pip")
        np.savetxt(f"{dir}/pip/{col}_sig_genes.txt", sig_genes, fmt="%s")
        print(f"Saved {len(sig_genes)} sig. genes for {col} (PIP > 0.9) in {dir}/pip/{col}_sig_genes.txt")
    skip_lfsr = os.path.exists(lfsr_path)

    if skip_lfsr:
        print("\nlfsr.csv exists. skipping lfsr compute...")
        lfsr_df = pd.read_csv(lfsr_path, index_col=0)
        lfsr_df.index = pd.Index(genes)
        lfsr_df.columns = perturbed
        lfsr_df.to_csv(lfsr_path)
    else:
        print("\ncomputing lfsr...")
        lfsr = compute_lfsr(key=rdm.PRNGKey(0), params=params, iters=2000)
        lfsr.block_until_ready()
        lfsr_np = np.array(lfsr)
        # lfsr_df = pd.DataFrame(lfsr_np.T)
        lfsr_df = pd.DataFrame(lfsr_np.T, index=genes, columns=perturbed)
        lfsr_df.to_csv(lfsr_path)

    # number of degs per w from PIP (find top genes with high pip)
    # top_pip_degs = find_top_genes(pip_df,0.90)
    # number of degs per w from PIP
    num_deg_per_w = (pip_df>0.9).sum(axis=0)


    # number of degs per perturbed gene from LFSR
    num_deg_per_gene = (lfsr_df<0.05).sum(axis=0)

    for col in lfsr_df.columns:
        sig_genes = lfsr_df.index[lfsr_df[col] < 0.05].tolist()
        if not os.path.exists(f"{dir}/lfsr"):
            os.makedirs(f"{dir}/lfsr")
        np.savetxt(f"{dir}/lfsr/{col}_sig_genes.txt", sig_genes, fmt="%s")
        print(f"Saved {len(sig_genes)} sig. genes for {col} (LFSR < 0.05) in {dir}/lfsr.")

    # p_hat_df = pd.DataFrame(params.p_hat.T, columns=column_names_b)
    p_hat_df = pd.DataFrame(params.p_hat.T, columns=column_names_b, index=perturbed)
    p_hat_df.to_csv(p_hat_path)
        
    # beta_df = pd.DataFrame(beta_sparse, columns=column_names_b)
    beta_df = pd.DataFrame(beta_sparse, columns=column_names_b, index=perturbed)
    beta_df.to_csv(beta_path)

    # overall_df = pd.DataFrame(overall_effect.T)
    overall_df = pd.DataFrame(overall_effect.T, columns=perturbed, index=genes)
    overall_df.to_csv(overall_path)

    print("\nshape of W df", W.shape)
    print("shape of pip df", pip_df.shape)
    print("shape of lfsr df", lfsr_df.shape)
    print("shape of p_hat df", p_hat_df.shape)
    print("shape of beta target df", beta_df.shape)
    print("shape of overall effect df", overall_df.shape)

    print("\nDone!\n")

    print(f"To check significant DEGs per W, see {dir}/pip")
    print(f"To check significant DEGs per perturbed gene, see {dir}/lfsr\n")

    return {
        "params": params,
        "W_df": W,
        "pip_df": pip_df,
        "pip_analysis": analysis,
        "num_deg_per_perturbed_gene": num_deg_per_gene,
        "num_deg_per_w": num_deg_per_w,
        "lfsr_df": lfsr_df,
        "p_hat_df": p_hat_df,
        "beta_df": beta_df,
        "overall_effect_df": overall_df,
    }



from typing import NamedTuple

import equinox as eqx

from jax import lax as lax, nn as nn, numpy as jnp
from jaxtyping import Array

from .common import DataMatrix, ModelParams
from .guide import GuideModel
from .utils import kl_discrete, logdet


class FactorParams(NamedTuple):
    mean_z: Array
    covar_z: Array


class FactorMoments(NamedTuple):
    mean_z: Array
    mean_zz: Array


class FactorModel(eqx.Module):

    def update(self, data: DataMatrix, guide: GuideModel, loadings: "LoadingModel", params: ModelParams) -> ModelParams:
        mean_w, mean_ww = loadings.moments(params)
        z_dim = params.z_dim

        update_var_z = jnp.linalg.inv(params.tau * mean_ww + jnp.identity(z_dim))
        update_mu_z = (params.tau * (data @ mean_w.T) + guide.predict(params)) @ update_var_z

        return params._replace(
            mean_z=update_mu_z,
            var_z=update_var_z,
        )

    def moments(self, params: ModelParams) -> FactorMoments:
        n_dim = params.n_dim

        # compute expected residuals
        # use posterior mean of Z, W, and Alpha to calculate residuals
        mean_z = params.mean_z
        mean_zz = mean_z.T @ mean_z + n_dim * params.var_z

        moments_ = FactorMoments(
            mean_z=mean_z,
            mean_zz=mean_zz,
        )
        return moments_

    def kl_divergence(self, guide: GuideModel, params: ModelParams) -> Array:
        n_dim, z_dim = params.n_dim, params.z_dim
        mean_z, var_z = params.mean_z, params.var_z
        pred_z = guide.predict(params)
        # tr(mean_zz) = tr(mean_z' mean_z) + tr(n * var_z)
        #  = sum(mean_z ** 2) + n * tr(var_z)
        # NB: tr(E_q[Z]' M E_prior[Z]) = sum(E_q[Z] * (M E_prior[Z])); saves factor of n
        # guide.weighted_sumsq(params) = tr(M'E[BB']M); can change depending on guide model
        t1 = jnp.sum(mean_z**2)
        t2 = n_dim * jnp.trace(var_z)
        t3 = -2 * jnp.sum(mean_z * pred_z)
        t4 = guide.weighted_sumsq(params)
        t5 = -n_dim * z_dim
        t6 = -n_dim * logdet(params.var_z)
        kl_d_ = 0.5 * (t1 + t2 + t3 + t4 + t5 + t6)
        return kl_d_


class LoadingMoments(NamedTuple):
    mean_w: Array
    mean_ww: Array


def _log_bf_np(z: Array, s2: Array, s0: Array) -> Array:
    s0_inv = 1.0 / s0
    s2ps0inv = s2 + s0_inv
    return 0.5 * (jnp.log(s2) - jnp.log(s2ps0inv) + z**2 * (s0_inv / s2ps0inv))


class _EffectLoopResults(NamedTuple):
    E_zzk: Array
    RtZk: Array
    Wk: Array
    k: int
    params: ModelParams


def _update_susie_effect(ldx: int, effect_params: _EffectLoopResults) -> _EffectLoopResults:
    E_zzk, RtZk, Wk, kdx, params = effect_params

    # remove current kl'th effect and update its expected residual
    Wkl = Wk - (params.mean_w[ldx, kdx] * params.alpha[ldx, kdx])
    E_RtZk = RtZk - E_zzk * Wkl

    # calculate update_var_w as the new V[w | gamma]
    # suppose indep between w_k
    update_var_wkl = jnp.reciprocal(params.tau * E_zzk + params.tau_0[ldx, kdx])

    # calculate update_mu_w as the new E[w | gamma]
    update_mean_wkl = params.tau * update_var_wkl * E_RtZk

    Z_s = (E_RtZk / E_zzk) * jnp.sqrt(E_zzk * params.tau)
    s2_s = 1 / (E_zzk * params.tau)
    s20_s = params.tau_0[ldx, kdx]
    log_bf = _log_bf_np(Z_s, s2_s, s20_s)
    # notice that pi is 1-D array for model without annotation
    # changed in the init_params
    log_alpha = jnp.log(params.pi[kdx, :]) + log_bf
    alpha_kl = nn.softmax(log_alpha)

    # update marginal w_kl
    Wk = Wkl + (update_mean_wkl * alpha_kl)
    params = params._replace(
        mean_w=params.mean_w.at[ldx, kdx].set(update_mean_wkl),
        var_w=params.var_w.at[ldx, kdx].set(update_var_wkl),
        alpha=params.alpha.at[ldx, kdx].set(alpha_kl),
    )

    return effect_params._replace(Wk=Wk, params=params)


class _FactorLoopResults(NamedTuple):
    X: DataMatrix
    W: Array
    EZZ: Array
    params: ModelParams


def _loop_factors(kdx: int, loop_params: _FactorLoopResults) -> _FactorLoopResults:
    data, W, mean_zz, params = loop_params
    l_dim, z_dim, _ = params.mean_w.shape

    # sufficient stats for inferring downstream w_kl/alpha_kl
    not_kdx = jnp.where(jnp.arange(z_dim) != kdx, size=z_dim - 1)
    E_zpzk = mean_zz[kdx][not_kdx]
    E_zzk = mean_zz[kdx, kdx]
    Wk = W[kdx, :]
    Wnk = W[not_kdx]
    RtZk = params.mean_z[:, kdx] @ data - Wnk.T @ E_zpzk

    # update over each of L effects
    init_loop_param = _EffectLoopResults(E_zzk, RtZk, Wk, kdx, params)
    _, _, Wk, _, params = lax.fori_loop(
        0,
        l_dim,
        _update_susie_effect,
        init_loop_param,
        unroll=False,
    )

    return loop_params._replace(W=W.at[kdx].set(Wk), params=params)


class LoadingModel(eqx.Module):

    def update(self, data: DataMatrix, factors: FactorModel, params: ModelParams) -> ModelParams:
        z_dim = params.z_dim
        _, mean_zz = factors.moments(params)
        mean_w, _ = self.moments(params)

        # update locals (W, alpha)
        init_loop_param = _FactorLoopResults(data, mean_w, mean_zz, params)
        _, _, _, params = lax.fori_loop(0, z_dim, _loop_factors, init_loop_param, unroll=False)
        return params

    @staticmethod
    def update_hyperparam(params: ModelParams) -> ModelParams:
        est_varw = params.mean_w**2 + params.var_w[:, :, jnp.newaxis]

        u_tau_0 = jnp.sum(params.alpha, axis=-1) / jnp.sum(est_varw * params.alpha, axis=-1)

        return params._replace(tau_0=u_tau_0)

    def moments(self, params: ModelParams) -> LoadingMoments:
        trace_var = jnp.sum(params.var_w[:, :, jnp.newaxis] * params.alpha, axis=(-1, 0))
        mu_w = jnp.sum(params.mean_w * params.alpha, axis=0)
        moments_ = LoadingMoments(
            mean_w=mu_w,
            mean_ww=mu_w @ mu_w.T + jnp.diag(trace_var),
        )

        return moments_

    @staticmethod
    def kl_divergence(params: ModelParams) -> Array:
        # technically this depends on the annotation model, but we usually flatten its predictions into the `pi`
        # member of `params`

        # awkward indexing to get broadcast working
        # KL for W variables
        klw_term1 = params.tau_0[:, :, jnp.newaxis] * (params.var_w[:, :, jnp.newaxis] + params.mean_w**2)
        klw_term2 = klw_term1 - 1.0 - (jnp.log(params.tau_0) + jnp.log(params.var_w))[:, :, jnp.newaxis]

        # weighted KL by E_q[gamma] variables
        kl_w_ = 0.5 * jnp.sum(params.alpha * klw_term2)

        # KL for gamma variables
        kl_gamma_ = kl_discrete(params.alpha, params.pi)

        return kl_w_ + kl_gamma_



from pathlib import Path
import argparse as ap
import logging
import sys
import os

import scanpy as sc
import numpy as np
import pandas as pd

import jax
import jax.numpy as jnp
from jax.experimental import sparse

from perturbvi.io import save_results
from perturbvi.log import get_logger
from perturbvi import infer

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_default_matmul_precision", "highest")

def main(args):
    argp = ap.ArgumentParser(description="Run perturbVI inference")
    argp.add_argument("matrix", type=str, help="perturbation matrix (csv) file")
    argp.add_argument("guide", type=str, help="guide csv file")
    argp.add_argument("z_dim", type=int, help="Number of latent factors")
    argp.add_argument("l_dim", type=int, help="Number of single effects")
    argp.add_argument("tau", type=int, help="residual precision")
    argp.add_argument("-o", "--output", type=str, help="Output directory")
    argp.add_argument(
        "--device", choices=["cpu", "gpu"], default="cpu", help="JAX device to use"
    )
    argp.add_argument(
        "--verbose", action="store_true", default=False, help="verbose logging",
    )

    args = argp.parse_args(args)

    out = args.output.rstrip('/')
    os.makedirs(out, exist_ok=True)
    log = get_logger(__name__, out)
    log.setLevel(logging.DEBUG if args.verbose else logging.INFO)

    matrix_path = Path(args.matrix)
    guide_path = Path(args.guide)
    ext = matrix_path.suffix.lower()

    if matrix_path.exists() and guide_path.exists():
        log.info("files OK!")
    else:
        log.error("files not found!")
        sys.exit(1)

    if ext == ".h5ad":
        dt = sc.read_h5ad(matrix_path)
        data = jnp.asarray(dt.X)
    elif ext == ".csv":
        data = jnp.asarray(pd.read_csv(matrix_path, index_col=0)).astype(jnp.float64)
    else:
        raise ValueError(f"Unsupported file type: {ext}")

    df_G = pd.read_csv(guide_path, index_col=0)
    df_G = df_G.drop(
        ["cell_barcode", "non-targeting", "Nontargeting"], axis=1, errors="ignore"
    )
    G = jnp.asarray(df_G).astype(jnp.float64)
    g_sp = sparse.bcoo_fromdense(G)
    del G, df_G

    log.info("starting inference...")

    results = infer(
        data,
        z_dim=args.z_dim,
        l_dim=args.l_dim,
        G=g_sp,
        A=None,
        p_prior=0.1,
        standardize=True,
        init="random",
        tau=args.tau,
        tol=1e-2,
        max_iter=500,
    )

    log.info("finished inference!")

    log.info(
        f"PVE across {args.z_dim} factors are {results.pve}; total PVE is {np.sum(results.pve)}"
    )

    save_results(results, path=out)
    log.info("saved results!")


def run_cli():
    return main(sys.argv[1:])

if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))


from importlib.metadata import PackageNotFoundError, version  # pragma: no cover

# annoying lint bug...
from .infer import (
    compute_elbo as compute_elbo,
    compute_pip as compute_pip,
    compute_pve as compute_pve,
    infer as infer,
)
from .sim import generate_sim as generate_sim
from .utils import (
    compute_lfsr as compute_lfsr,
    luhmes_analysis as luhmes_analysis,
)
from .io import save_results as save_results


try:
    # Change here if project is renamed and does not equal the package name
    dist_name = __name__
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError



import pickle

from typing import Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from adjustText import adjust_text

from .infer import InferResults


__all__ = ["save_results"]


# save all results as a pickle object
def save_results(results: InferResults, path: str):
    """Create a function to save SuSiE PCA results returned by function
    perturbvi.infer

    Args:
        results: results object returned by perturbvi.infer
        path: local path to save the results subject


    """
    print("Save results from SuSiE PCA")

    np.savetxt(f"{path}/W.txt", results.W)
    np.savetxt(f"{path}/pip.txt", results.pip)
    np.savetxt(f"{path}/pve.txt", results.pve)

    params_file = open(f"{path}/params_file.pkl", "wb")
    pickle.dump(results.params, params_file)
    params_file.close()

    print(f"Results saved successfully at {path}")

    return


# # function to find genes with high pip
# def find_top_genes(
#     results: InferResults, pip_cutoff: float = 0.9, gene_symbol: Optional[list] = None, filepath: Optional[str] = None
# ) -> dict:
#     """Find genes with high posterior inclusion probabilities (PIPs) and optionally save results.

#     Args:
#         results: InferResults object containing inference results
#         pip_cutoff: Threshold for considering a PIP value as significant (default: 0.9)
#         gene_symbol: Optional list of gene symbols to use as row indices. If None, uses numeric indices
#         filepath: Optional path to save results. If provided, saves results as CSV with columns:
#                  'factor', 'gene', 'pip_value'

#     Returns:
#         dict: Dictionary mapping each factor to list of genes with PIP > cutoff
#     """
#     # Create PIP DataFrame
#     z_dim, p_dim = results.params.W.shape
#     column_names = [f"w{i}" for i in range(z_dim)]

#     if gene_symbol is not None:
#         if len(gene_symbol) != p_dim:
#             raise ValueError(f"Length of gene_symbol ({len(gene_symbol)}) must match number of genes ({p_dim})")
#         df = pd.DataFrame(results.pip.T, columns=column_names, index=gene_symbol)
#     else:
#         df = pd.DataFrame(results.pip.T, columns=column_names)

#     # Initialize an empty dictionary
#     high_value_genes = {}

#     # Iterate over each column (factor) in the DataFrame
#     for column in df.columns:
#         # Find the rows in the current column with values greater than cutoff
#         high_values_mask = df[column] > pip_cutoff
#         high_values = df.index[high_values_mask].tolist()
#         # Store these rows in the dictionary
#         high_value_genes[column] = high_values

#     # Save to file if filepath is provided
#     if filepath is not None:
#         # Create a list to store rows for the DataFrame
#         rows = []
#         for factor, genes in high_value_genes.items():
#             for gene in genes:
#                 rows.append({"factor": factor, "gene": gene, "pip_value": df.loc[gene, factor]})
#         # Convert to DataFrame and save
#         results_df = pd.DataFrame(rows)
#         results_df.to_csv(filepath, index=False)

#     return high_value_genes


def save_pip(results: InferResults, gene_symbol: Optional[list] = None, filepath: Optional[str] = None) -> pd.DataFrame:
    """Create a DataFrame of posterior inclusion probabilities (PIPs) and optionally save to CSV.

    Args:
        results: InferResults object containing inference results
        gene_symbol: Optional list of gene symbols to use as row indices. If None, uses numeric indices
        filepath: Optional path to save CSV file. If provided, saves DataFrame to this location

    Returns:
        pd.DataFrame: DataFrame containing PIP values with labeled columns and indices
    """
    z_dim, p_dim = results.params.W.shape
    # Create column names in format 'w0', 'w1', etc.
    column_names = [f"w{i}" for i in range(z_dim)]

    # Create DataFrame with appropriate indices
    if gene_symbol is not None:
        if len(gene_symbol) != p_dim:
            raise ValueError(f"Length of gene_symbol ({len(gene_symbol)}) must match number of genes ({p_dim})")
        pip_df = pd.DataFrame(results.pip.T, columns=column_names, index=gene_symbol)
    else:
        pip_df = pd.DataFrame(results.pip.T, columns=column_names)

    # Save to CSV if filepath is provided
    if filepath is not None:
        pip_df.to_csv(filepath)

    return pip_df


# plot beta
def plot_beta(
    results: InferResults,
    factor_idx: int,
    perturb_names: Optional[list] = None,
    n: int = 20,
    figsize: Tuple[int, int] = (10, 6),
    fontsize: int = 8,
    save_path: Optional[str] = None,
    dpi: int = 300,
) -> None:
    """Plot beta values for a specific factor from inference results.
    Highlights the top n points with the highest absolute values.
    Labels are added near the top n data points, with automatic adjustment to avoid overlap.

    Args:
        results: InferResults object containing inference results
        factor_idx: Index of the factor to plot (0-based)
        perturb_names: Optional list of perturbation names to use as labels. If None, uses numeric indices
        n: Number of top values to highlight and label (default: 20)
        figsize: Figure size as (width, height) tuple (default: (10, 6))
        fontsize: Font size for gene labels (default: 8)
        save_path: Optional path to save the plot as PNG file. If None, displays plot instead
        dpi: Resolution of the saved image in dots per inch (default: 300)
    """
    # Extract parameters and compute sparse beta matrix
    params = results.params
    z_dim = params.mean_beta.shape[1]

    # Create beta matrix and dataframe
    beta_sparse = params.mean_beta * params.p_hat.T
    column_names = [f"b{i}" for i in range(z_dim)]
    df = pd.DataFrame(beta_sparse, columns=column_names)

    if perturb_names is not None:
        if len(perturb_names) != len(df):
            raise ValueError(
                f"Length of perturb_names ({len(perturb_names)}) must match number of perturbations ({len(df)})"
            )
        df.index = perturb_names

    column = f"b{factor_idx}"
    if column not in df.columns:
        raise ValueError(f"Factor index {factor_idx} out of range. Should be between 0 and {z_dim-1}")

    plt.figure(figsize=figsize)

    # Numeric mapping for string indices
    numeric_indices = range(len(df))

    top_values = df[column].abs().nlargest(n)
    top_indices = top_values.index
    # draw the horizontal line
    top_value = df[column].abs().nlargest(n)[-1]

    # Use numeric indices for plotting
    plt.scatter(numeric_indices, df[column], color="grey", alpha=0.7)
    plt.scatter(
        [numeric_indices[df.index.get_loc(i)] for i in top_indices], df.loc[top_indices, column], color="red", zorder=5
    )

    plt.axhline(y=top_value, color="green", linestyle="--")
    plt.axhline(y=0, color="black", linestyle="-")
    plt.axhline(y=0 - top_value, color="green", linestyle="--")

    texts = []
    for index in top_indices:
        x_pos = numeric_indices[df.index.get_loc(index)]
        y_pos = df.loc[index, column]
        # Use perturbation name if provided, otherwise use index
        label = str(index)
        texts.append(plt.text(x_pos, y_pos, label, fontsize=fontsize, color="red"))

    adjust_text(texts, arrowprops=dict(arrowstyle="-", color="red"))

    plt.xticks([])
    plt.ylabel("Beta Value")
    plt.title(f"Top {n} Beta Values for Factor {factor_idx}")

    if save_path is not None:
        plt.savefig(save_path, dpi=dpi, bbox_inches="tight")
        plt.close()
    else:
        plt.show()


def draw_perturb_heatmap(results: InferResults, perturb_names: Optional[list] = None,
                        figsize: Tuple[int, int] = (10, 6), cmap: str = 'seismic',
                        save_path: Optional[str] = None, dpi: int = 300) -> None:
    """Draw heatmap of perturbation effects across factors.
    
    Args:
        results: InferResults object containing inference results
        perturb_names: Optional list of perturbation names for y-axis labels
        figsize: Figure size as (width, height) tuple (default: (10, 6))
        cmap: Color map for heatmap (default: 'seismic')
        save_path: Optional path to save the plot as PNG file. If None, displays plot instead
        dpi: Resolution of the saved image in dots per inch (default: 300)
    """
    # Extract parameters and compute sparse beta matrix
    params = results.params
    z_dim = params.mean_beta.shape[1]
    
    # Create beta matrix and dataframe
    beta_sparse = params.mean_beta * params.p_hat.T
    column_names = [f'{i}' for i in range(z_dim)]
    df = pd.DataFrame(beta_sparse, columns=column_names)
    
    if perturb_names is not None:
        if len(perturb_names) != len(df):
            raise ValueError(f"Length of perturb_names ({len(perturb_names)}) must match number of perturbations ({len(df)})")
        df.index = perturb_names

    # Create figure and axes
    fig, ax = plt.subplots(figsize=figsize)
    
    # Create symmetric color scale
    vmax = np.max(np.abs(df))
    
    # Draw heatmap
    sns.heatmap(df, cmap=cmap, center=0, vmin=-vmax, vmax=vmax, ax=ax)
    
    # Move x-axis labels and ticks to top
    ax.xaxis.tick_top()
    ax.xaxis.set_label_position('top')
    
    # Set labels
    ax.set_xlabel("Factors")
    ax.set_ylabel("Perturbations")
    
    if save_path is not None:
        plt.savefig(save_path, dpi=dpi, bbox_inches='tight')
        plt.close()
    else:
        plt.show()




from abc import abstractmethod
from dataclasses import field
from typing import Any

import equinox as eqx
import jax
import jax.nn as nn
import optimistix as optx

from jaxtyping import Array

from .common import DataMatrix, ModelParams
from .utils import kl_discrete


def _compute_pi(A: DataMatrix, theta: Array) -> Array:
    return nn.softmax(A @ theta, axis=0).T


def _loss(theta: Array, args) -> Array:
    A, alpha = args
    pi = _compute_pi(A, theta)
    return kl_discrete(alpha, pi), None


class PriorModel(eqx.Module):
    @abstractmethod
    def predict(self, params: ModelParams) -> Array: ...

    @abstractmethod
    def init_state(self, params: ModelParams) -> ModelParams: ...

    @abstractmethod
    def update(self, params: ModelParams) -> ModelParams: ...


class FixedPrior(PriorModel):
    def predict(self, params: ModelParams) -> Array:
        return params.pi

    def init_state(self, params: ModelParams) -> ModelParams:
        return params

    def update(self, params: ModelParams) -> ModelParams:
        return params


class AnnotationPriorModel(PriorModel):
    A: DataMatrix
    search: optx.AbstractMinimiser
    step: Any = field(init=False)

    def __post_init__(self):
        self.step = eqx.filter_jit(eqx.Partial(self.search.step, _loss, options=None, tags=None))

    @property
    def shape(self):
        return self.A.shape

    def init_state(self, params: ModelParams) -> ModelParams:
        args = (self.A, params.alpha)
        f_struct, aux_struct = jax.eval_shape(_loss, params.theta, args)
        return params._replace(
            ann_state=self.search.init(
                _loss, params.theta, args, options=None, f_struct=f_struct, aux_struct=aux_struct, tags=None
            ),
        )

    def predict(self, params: ModelParams) -> Array:
        return _compute_pi(self.A, params.theta)

    def update(self, params: ModelParams) -> ModelParams:
        args = (self.A, params.alpha)
        # take one step using optimistix optimizer
        theta, state, _ = self.step(params.theta, state=params.ann_state, args=args)
        return params._replace(
            theta=theta,
            pi=_compute_pi(self.A, theta),
            ann_state=state,
        )



from typing import get_args, Literal, NamedTuple, Optional, Tuple

from plum import dispatch

import equinox as eqx
import jax
import optax
import optimistix as optx

from jax import jit, nn, numpy as jnp, random
from jax.experimental import sparse
from jaxtyping import Array, ArrayLike

from .annotation import AnnotationPriorModel, FixedPrior, PriorModel
from .common import (
    DataMatrix,
    ModelParams,
)
from .factorloadings import FactorModel, LoadingModel
from .guide import DenseGuideModel, GuideModel, SparseGuideModel
from .sparse import CenteredSparseMatrix, SparseMatrix
from .utils import prob_pca


_init_type = Literal["pca", "random"]


@dispatch
def _is_valid(X: ArrayLike):
    return jnp.all(jnp.isfinite(X))


@dispatch
def _is_valid(X: sparse.JAXSparse):
    return jnp.all(jnp.isfinite(X.data))


def _update_tau(X: DataMatrix, factor: FactorModel, loadings: LoadingModel, params: ModelParams) -> ModelParams:
    n_dim, p_dim = X.shape

    # calculate moments of factors and loadings
    mean_z, mean_zz = factor.moments(params)
    mean_w, mean_ww = loadings.moments(params)

    # expectation of log likelihood
    # tr(A @ B) == sum(A * B)
    E_ss = params.x_ssq - 2 * jnp.trace(mean_w @ (X.T @ mean_z)) + jnp.sum(mean_zz * mean_ww)
    u_tau = (n_dim / E_ss) * p_dim

    return params._replace(tau=u_tau)


class ELBOResults(NamedTuple):
    """Define the class of all components in ELBO.

    **Arguments:**
        elbo: the value of ELBO
        expected_loglike: Expectation of log-likelihood
        kl_factors: -KL divergence of Z
        kl_loadings: -KL divergence of W
        kl_guide: -KL divergence of B

    """

    elbo: Array
    expected_loglike: Array
    kl_factors: Array
    kl_loadings: Array
    kl_guide: Array

    def __str__(self):
        return (
            f"ELBO = {self.elbo:.3f} | E[logl] = {self.expected_loglike:.3f} | "
            f"KL[Z] = {self.kl_factors:.3f} | E_Q[KL[W]] + KL[Gamma] = {self.kl_loadings:.3f} | "
            f"E_Q[KL[Beta]] + KL[Eta] = {self.kl_guide:.3f}|"
        )


def compute_elbo(
    X: DataMatrix,
    guide: GuideModel,
    factors: FactorModel,
    loadings: LoadingModel,
    annotation: PriorModel,
    params: ModelParams,
) -> ELBOResults:
    """Create function to compute evidence lower bound (ELBO)

    **Arguments:**

    - `X` [`Array`]: The observed data, an N by P ndarray
    - `guide` [`GuideModel`]: The guide model
    - `factors` [`FactorModel`]: The factor model
    - `loadings` [`LoadingModel`]: The loading model
    - `annotation` [`PriorModel`]: The prior annotation model
    - `params` [`ModelParams`]: The dictionary contains all the inferred parameters

    **Returns:**
    - `ELBOResults` [`ELBOResults`]: The object contains all components in ELBO

    """
    n_dim, _ = params.mean_z.shape
    _, _, p_dim = params.mean_w.shape

    # calculate second moment of Z along k, (k x k) matrix
    # E[Z'Z] = V_k[Z] * tr(I_n) + E[Z]'E[Z] = V_k[Z] * n + E[Z]'E[Z]
    mean_z, mean_zz = factors.moments(params)

    # calculate moment of W
    mean_w, mean_ww = loadings.moments(params)

    # expectation of log likelihood
    # calculation tip: tr(A @ A.T) = tr(A.T @ A) = sum(A ** 2); but doubles mem
    # tr(A.T @ A) is inner product of A with itself = vdot(X, X)
    # (X.T @ E[Z] @ E[W]) is p x p (big!); compute (E[W] @ X.T @ E[Z]) (k x k)
    # jnp.vdot(X, X) is const wrt ELBO but costly to compute; could either compute once and store
    # or just ignore it
    exp_logl = (-0.5 * params.tau) * (
        params.x_ssq
        -2 * jnp.einsum("kp,np,nk->", mean_w, X, mean_z)  # tr(E[W] @ X.T @ E[Z])
        + jnp.einsum("ij,ji->", mean_zz, mean_ww)  # tr(E[Z.T @ Z] @ E[W @ W.T])
    ) + 0.5 * n_dim * p_dim * jnp.log(params.tau)

    # neg-KL for Z
    kl_factors = factors.kl_divergence(guide, params)

    # neg-KL for w
    kl_loadings = loadings.kl_divergence(params)

    # neg-KL for beta
    kl_guide = guide.kl_divergence(params)

    elbo = exp_logl - (kl_factors + kl_loadings + kl_guide)

    result = ELBOResults(elbo, exp_logl, kl_factors, kl_loadings, kl_guide)

    return result


@eqx.filter_jit
def _inner_loop(
    X: DataMatrix,
    guide: GuideModel,
    factors: FactorModel,
    loadings: LoadingModel,
    annotation: PriorModel,
    params: ModelParams,
):
    # update annotation priors if any
    params = annotation.init_state(params)
    params = annotation.update(params)

    # update loadings prior precision via ~Empirical Bayes and update variational params
    params = loadings.update_hyperparam(params)
    params = loadings.update(X, factors, params)

    # update factor parameters
    params = factors.update(X, guide, loadings, params)

    # update beta and p_hat
    params = guide.update_hyperparam(params)
    params = guide.update(params)

    # update precision parameters via MLE
    params = _update_tau(X, factors, loadings, params)

    # compute elbo
    elbo_res = compute_elbo(X, guide, factors, loadings, annotation, params)

    return elbo_res, params


def _reorder_factors_by_pve(pve: Array, annotations: PriorModel, params: ModelParams) -> Tuple[Array, ModelParams]:
    sorted_indices = jnp.argsort(pve)[::-1]
    pve = pve[sorted_indices]

    sorted_mu_z = params.mean_z[:, sorted_indices]
    sorted_var_z = params.var_z[sorted_indices, sorted_indices]
    sorted_mu_beta = params.mean_beta[:, sorted_indices]
    sorted_var_beta = params.var_beta[:, sorted_indices]
    sorted_p_hat = params.p_hat[sorted_indices, :]
    sorted_mu_w = params.mean_w[:, sorted_indices, :]
    sorted_var_w = params.var_w[:, sorted_indices]
    sorted_tau_beta = params.tau_beta[sorted_indices]
    sorted_alpha = params.alpha[:, sorted_indices, :]
    sorted_tau_0 = params.tau_0[:, sorted_indices]
    if isinstance(annotations, AnnotationPriorModel):
        sorted_theta = params.theta[:, sorted_indices]
        # sorted_pi = annotations.predict(ModelParams(theta=sorted_theta))  # type: ignore
        sorted_pi = annotations.predict(params._replace(theta=sorted_theta))
    else:
        sorted_theta = None
        sorted_pi = params.pi

    params = ModelParams(
        params.x_ssq,
        sorted_mu_z,
        sorted_var_z,
        sorted_mu_w,
        sorted_var_w,
        sorted_alpha,
        params.tau,
        sorted_tau_0,
        sorted_theta,
        sorted_pi,
        None,
        sorted_mu_beta,
        sorted_var_beta,
        sorted_tau_beta,
        params.p,
        sorted_p_hat,
    )

    return pve, params


def _init_params(
    rng_key: random.PRNGKey,
    z_dim: int,
    l_dim: int,
    X: DataMatrix,
    guide: GuideModel,
    factors: FactorModel,
    loadings: LoadingModel,
    annotations: PriorModel,
    p_prior: float = 0.5,
    tau: float = 1.0,
    init: _init_type = "pca",
) -> ModelParams:
    print("Starting model parameter initialization...")
    
    # Base parameters
    n_dim, p_dim = X.shape
    tau_0 = jnp.ones((l_dim, z_dim))
    tau_0.block_until_ready()
    print("✓ Base parameters initialized (5%)")

    # Random keys
    keys = random.split(rng_key, 10)
    keys[0].block_until_ready()
    rng_key, svd_key, mu_key, var_key, muw_key, varw_key, alpha_key, beta_key, var_beta_key, theta_key = keys
    print("✓ Random keys setup (10%)")

    # Data statistics
    x_ssq = jnp.sum(X * X)
    x_ssq.block_until_ready()
    print("✓ Data statistics computed (15%)")

    # Factors
    if init == "pca":
        init_mu_z, _ = prob_pca(svd_key, X, k=z_dim)
    else:
        init_mu_z = random.normal(mu_key, shape=(n_dim, z_dim))
    init_mu_z.block_until_ready()
    print("✓ Factors initialized (35%)")

    # Factor variance
    init_var_z = jnp.diag(random.normal(var_key, shape=(z_dim,)) ** 2)
    init_var_z.block_until_ready()
    print("✓ Factor variance set (45%)")

    # Loadings
    init_mu_w = random.normal(muw_key, shape=(l_dim, z_dim, p_dim)) * 1e-3
    init_var_w = (1 / tau_0) * (random.normal(varw_key, shape=(l_dim, z_dim))) ** 2
    init_mu_w.block_until_ready()
    init_var_w.block_until_ready()
    print("✓ Loadings initialized (60%)")

    # Alpha and pi
    #init_alpha = random.dirichlet(alpha_key, alpha=jnp.ones(p_dim), shape=(l_dim, z_dim))
    #init_alpha =jax.jit(random.dirichlet)(alpha_key, alpha=jnp.ones(p_dim), shape=(l_dim, z_dim))
    init_alpha = jnp.full((l_dim, z_dim, p_dim), 1. / p_dim)
    print("Avoid dirichlet process")
    init_alpha.block_until_ready()
    if isinstance(annotations, AnnotationPriorModel):
        p_dim, m = annotations.shape
        theta = random.normal(theta_key, shape=(m, z_dim))
        pi = nn.softmax(annotations.A @ theta, axis=0).T
        theta.block_until_ready()
        pi.block_until_ready()
    else:
        theta = None
        pi = jnp.ones(shape=(z_dim, p_dim)) / p_dim
        pi.block_until_ready()
    print("✓ Annotations setup complete (75%)")

    # Perturbation effects
    n_dim, g_dim = guide.shape
    tau_beta = jnp.ones((z_dim,))
    init_mu_beta = random.normal(beta_key, shape=(g_dim, z_dim)) * 1e-3
    init_var_beta = (1 / tau_beta) * random.normal(var_beta_key, shape=(g_dim, z_dim)) ** 2
    tau_beta.block_until_ready()
    init_mu_beta.block_until_ready()
    init_var_beta.block_until_ready()
    print("✓ Perturbation effects initialized (90%)")

    # Priors
    if p_prior is not None:
        p_prior = p_prior * jnp.ones(g_dim)
        p_prior.block_until_ready()
    p_hat = 0.5 * jnp.ones(shape=(z_dim, g_dim))
    p_hat.block_until_ready()
    print("✓ Priors setup complete (100%)")
    
    print("\n✓ Model parameter initialization completed successfully")

    return ModelParams(
        x_ssq,
        init_mu_z,
        init_var_z,
        init_mu_w,
        init_var_w,
        init_alpha,
        jnp.array(tau, dtype=float),
        tau_0,
        theta=theta,
        pi=pi,
        ann_state=None,
        mean_beta=init_mu_beta,
        var_beta=init_var_beta,
        tau_beta=tau_beta,
        p=p_prior,
        p_hat=p_hat,
    )


def _check_args(
    X: ArrayLike | sparse.JAXSparse, A: Optional[ArrayLike | sparse.JAXSparse], z_dim: int, l_dim: int, init: _init_type
) -> Tuple[Array | sparse.JAXSparse, Array | sparse.JAXSparse]:
    # pull type options for init
    type_options = get_args(_init_type)

    if isinstance(X, ArrayLike):
        X = jnp.asarray(X)

    if X.ndim != 2:
        raise ValueError(f"Shape of X = {X.shape}; Expected 2-dim matrix")

    # should we check for n < p?
    n_dim, p_dim = X.shape

    # dim checks
    if l_dim > p_dim:
        raise ValueError(f"l_dim should be less than p: received l_dim = {l_dim}, p = {p_dim}")
    if l_dim <= 0:
        raise ValueError(f"l_dim should be positive: received l_dim = {l_dim}")
    if z_dim > p_dim:
        raise ValueError(f"z_dim should be less than p: received z_dim = {z_dim}, p = {p_dim}")
    if z_dim > n_dim:
        raise ValueError(f"z_dim should be less than n: received z_dim = {z_dim}, n = {n_dim}")
    if z_dim <= 0:
        raise ValueError(f"z_dim should be positive: received z_dim = {z_dim}")
    # quality checks
    if not _is_valid(X):
        raise ValueError("X contains 'nan/inf'. Please check input data for correctness or missingness")

    if A is not None:
        if isinstance(A, ArrayLike):
            A = jnp.asarray(A)
        if A.ndim != 2:
            raise ValueError(f"Dimension of annotation matrix A should be 2: received {len(A.shape)}")
        a_p_dim, _ = A.shape
        if a_p_dim != p_dim:
            raise ValueError(
                f"Leading dimension of annotation matrix A should match feature dimension {p_dim}: received {a_p_dim}"
            )
        if not _is_valid(A):
            raise ValueError("A contains 'nan/inf'. Please check input data for correctness or missingness")
    # type check for init

    if init not in type_options:
        raise ValueError(f"Unknown initialization provided '{init}'; Choices: {type_options}")

    return X, A  # type: ignore


class InferResults(NamedTuple):
    """Define the results object returned by function :py:obj:`susie_pca`.

    Attributes:
        params: the dictionary contain all the infered parameters
        elbo: the value of ELBO
        pve: the ndarray of percent of variance explained
        pip: the ndarray of posterior inclusion probabilities
        W: the posterior mean parameter for loadings

    """

    params: ModelParams
    elbo: ELBOResults
    pve: Array
    pip: Array

    @property
    def W(self) -> Array:
        return self.params.W


def infer(
    X: ArrayLike | sparse.JAXSparse,
    z_dim: int,
    l_dim: int,
    G: ArrayLike | sparse.JAXSparse,
    A: Optional[ArrayLike | sparse.JAXSparse] = None,
    p_prior: Optional[float] = 0.5,
    tau: float = 1.0,
    standardize: bool = False,
    init: _init_type = "pca",
    learning_rate: float = 1e-2,
    max_iter: int = 400,
    tol: float = 1e-3,
    seed: int = 0,
    verbose: bool = True,
) -> InferResults:
    """The main inference function for SuSiE PCA.

    **Arguments:**

    -`X` [`Array`|`JAXSparse`]: The expression count matrix. Should be an array-like or sparse JAX matrix.

    -`z_dim` [`int`]: The latent dimension.

    -`l_dim` [`int`]: The number of single effects in each factor.

    -`G` [`Array`|`JAXSparse`]: Perturbation density matrix. Should be an array-like or sparse JAX matrix.

    -`A` [`Array`]: Annotation matrix to use in parameterized-prior mode. If not `None`, leading dimension
        should match the feature dimension of X.

    -`p_prior` [`float`]: Prior probability for each perturbation to have a non-zero effect to predict latent factor.
        (default = 0.5)

    -`tau` [`float`]: initial value of residual precision (default = 1)

    -`standardize` [`bool`]: Whether to scale the input data with variance 1 (default = False)

    -`init` [`str`]: How to initialize the variational mean parameters for latent factors.
        Either "pca" or "random" (default = "pca").

    -`learning_rate` [`float`]: Learning rate for prior annotation probability inference. Not used if `A` is `None`.

    -`max_iter` [`int`]: Maximum number of iterations for inference.

    -`tol` [`float`]: Convergence tolerance for inference.

    -`seed` [`int`]: Numerical tolerance for ELBO convergence.

    -`verbose` [`bool`]: Flag to indicate displaying log information (ELBO value) in each
            iteration.

    **Returns:**

    An [`InferResults`][] object  contain all the inferred parameters.
    """

    # sanity check arguments
    X, A = _check_args(X, A, z_dim, l_dim, init)

    # cast to jax array
    if isinstance(X, Array):
        # option to center the data
        X -= jnp.mean(X, axis=0)
        if standardize:
            X /= jnp.std(X, axis=0)
    elif isinstance(X, sparse.JAXSparse):
        X = CenteredSparseMatrix(X, scale=standardize)  # type: ignore
    if isinstance(G, ArrayLike):
        G = jnp.asarray(G)
    elif isinstance(G, sparse.JAXSparse):
        G = SparseMatrix(G)

    if p_prior is None or jnp.isclose(p_prior, 0.0):
        guide = DenseGuideModel(G)
    else:
        guide = SparseGuideModel(G)

    if A is not None:
        adam = optax.adam(learning_rate)
        annotation = AnnotationPriorModel(A, optx.OptaxMinimiser(adam, rtol=1e-3, atol=1e-3))
    else:
        annotation = FixedPrior()

    n, p = X.shape  # type: ignore

    # initialize PRNGkey and params
    rng_key = random.PRNGKey(seed)
    factors = FactorModel()
    loadings = LoadingModel()
    params = _init_params(rng_key, z_dim, l_dim, X, guide, factors, loadings, annotation, p_prior, tau, init)

    #  core loop for inference
    elbo = -5e25
    elbo_res = None
    for idx in range(1, max_iter + 1):
        elbo_res, params = _inner_loop(X, guide, factors, loadings, annotation, params)

        if verbose:
            print(f"Iter [{idx}] | {elbo_res}")

        diff = elbo_res.elbo - elbo
        if diff < 0 and verbose:
            print(f"Alert! Diff between elbo[{idx - 1}] and elbo[{idx}] = {diff}")
        if jnp.fabs(diff) < tol:
            if verbose:
                print(f"Elbo diff tolerance reached at iteration {idx}")
            break

        elbo = elbo_res.elbo

    # compute PVE and reorder in descending value
    pve = compute_pve(params)
    pve, params = _reorder_factors_by_pve(pve, annotation, params)

    # compute PIPs
    pip = compute_pip(params)

    return InferResults(params, elbo_res, pve, pip)


def compute_pip(params: ModelParams) -> Array:
    """Compute the posterior inclusion probabilities (PIPs).

    **Arguments:**

    -`params` [`ModelParams`]: Instance of inferred parameters

    **Returns:**

    -`PIP` [`Array`]: Array of posterior inclusion probabilities (PIPs) for each of `K x P` factor,
              feature combinations

    """

    pip = -jnp.expm1(jnp.sum(jnp.log1p(-params.alpha), axis=0))

    return pip


def compute_pve(params: ModelParams) -> Array:
    """Compute the percent of variance explained (PVE).

    **Arguments:**

    -`params` [`ModelParams`]: Instance of inferred parameters

    **Returns:**

    -`PVE` [`Array`]: Array of length `K` that contains percent of variance
              explained by each factor (PVE)
    """

    n_dim = params.n_dim
    W = params.W
    z_dim, p_dim = W.shape

    sk = jnp.zeros(z_dim)
    for k in range(z_dim):
        sk = sk.at[k].set(jnp.sum((params.mean_z[:, k, jnp.newaxis] * W[k, :]) ** 2))

    s = jnp.sum(sk)
    pve = sk / (s + p_dim * n_dim * (1 / params.tau))

    return pve



from typing import NamedTuple, Optional

import numpy as np

import jax.numpy as jnp

from jax import random
from jax.typing import ArrayLike


__all__ = [
    "SimulatedData",
    "generate_sim",
]


class SimulatedData(NamedTuple):
    """the object contain simulated data components.

    Args:
        Z: simulated factor
        W: simulated loadings
        X: simulated data set
        G: design matrix
        beta: perturbation effect matrix

    """

    Z: ArrayLike
    W: ArrayLike
    X: ArrayLike
    G: Optional[ArrayLike]
    beta: Optional[ArrayLike]


# Define the function to generate design matrix.
# Ensure each perturbation is at least assigned to a cell
def create_design_matrix(key, n_dim, g_dim):
    """Create a design matrix given N and G

    Args:
    key : Random seed for np.random.
    n_dim : Sample size
    g_dim : Perturbation dimension

    Returns:
    G : The design matrix.
    """
    if g_dim > n_dim:
        raise ValueError(f"g_dim should be less than n: received g_dim = {g_dim}, n = {n_dim}")
    # Initialize the matrix with zeros
    # Assign each perturbation exactly once
    G = jnp.identity(g_dim)

    # Randomly assign remaining perturbations
    np.random.seed(key)
    indices = np.random.choice(g_dim, size=n_dim - g_dim)
    G_left = np.zeros((n_dim - g_dim, g_dim), dtype=int)
    G_left[np.arange(n_dim - g_dim), indices] = 1

    G = jnp.concatenate((G, G_left), axis=0)
    return G


def generate_sim(
    seed: int,
    l_dim: int,
    n_dim: int,
    p_dim: int,
    z_dim: int,
    g_dim: int,
    b_sparsity: float = 0.2,
    effect_size: float = 1.0,
) -> SimulatedData:
    """Create the function to generate a sparse data for perturbVI.
       Please make sure l_dim < p_dim/z_dim
       when generate simulation data using this function.

    Args:
        seed: Seed for random initialization
        l_dim: Number of single effects in each factor
        n_dim: Number of sample in the data
        p_dim: Number of feature in the data
        z_dim: Number of Latent dimensions
        g_dim: perturbation dimensions
        b_sparsity: sparsity of perturbation effects.
        effect_size: The effect size of features contributing to the factor.
                      (default = 1).

    Returns:
        SimulatedData: Tuple that contains simulated factors (`N x K`),
    """

    # interger seed
    if isinstance(seed, int) is False:
        raise ValueError(f"seed should be an interger: received seed = {seed}")

    rng_key = random.PRNGKey(seed)
    rng_key, b_key, beta_key, s_key, var_key, obs_key = random.split(rng_key, 6)

    # dimension check
    if l_dim > p_dim:
        raise ValueError(f"l_dim should be less than p: received l_dim = {l_dim}, p = {p_dim}")
    if l_dim > p_dim / z_dim:
        raise ValueError(
            f"""l_dim is smaller than p_dim/z_dim,
            please make sure each component has {l_dim} single effects"""
        )

    if l_dim <= 0:
        raise ValueError(f"l_dim should be positive: received l_dim = {l_dim}")

    if z_dim > p_dim:
        raise ValueError(f"z_dim should be less than p: received z_dim = {z_dim}, p = {p_dim}")
    if z_dim > n_dim:
        raise ValueError(f"z_dim should be less than n: received z_dim = {z_dim}, n = {n_dim}")
    if z_dim <= 0:
        raise ValueError(f"z_dim should be positive: received z_dim = {z_dim}")

    if effect_size <= 0:
        raise ValueError(f"effect size should be positive: received effect_size = {effect_size}")

    # random W
    W = jnp.zeros(shape=(z_dim, p_dim))

    for k in range(z_dim):
        W = W.at[k, (k * l_dim) : ((k + 1) * l_dim)].set(effect_size * random.normal(b_key, shape=(l_dim,)))

    # linear function to generate Z
    # perturbation effects
    beta = jnp.zeros(shape=(g_dim, z_dim))
    # number of non-zero entries in each component
    non_zero_num = int(b_sparsity * g_dim)

    for col in range(z_dim):
        beta_key, beta_key_2 = random.split(beta_key)
        # Step 2: Randomly select `non_zero_num` indices without replacement
        indices = random.choice(beta_key, g_dim, shape=(non_zero_num,), replace=False)

        # Step 3: Generate `non_zero_num` random values from a normal distribution
        random_values = random.normal(beta_key_2, (non_zero_num,))

        # Update the selected entries in the column with random values
        beta = beta.at[indices, col].set(random_values)

    G = create_design_matrix(s_key, n_dim, g_dim)

    Z = G @ beta + random.normal(var_key, shape=(n_dim, z_dim))
    # Latent factor model
    m = Z @ W
    X = m + random.normal(obs_key, shape=(n_dim, p_dim))

    return SimulatedData(Z, W, X, G, beta)


def generate_sim_with_control(
    seed: int,
    l_dim: int,
    n_dim: int,
    p_dim: int,
    z_dim: int,
    g_dim: int,
    control_fraction: float = 0.2,
    b_sparsity: float = 0.2,
    effect_size: float = 1.0,
) -> SimulatedData:
    """Create the function to generate a sparse data for perturbVI.
       Please make sure l_dim < p_dim/z_dim
       when generate simulation data using this function.

    Args:
        seed: Seed for random initialization
        l_dim: Number of single effects in each factor
        n_dim: Number of sample in the data
        p_dim: Number of feature in the data
        z_dim: Number of Latent dimensions
        g_dim: perturbation dimensions
        control_fraction: fraction of negative control guide
        b_sparsity: sparsity of perturbation effects.
        effect_size: The effect size of features contributing to the factor.
                      (default = 1).

    Returns:
        SimulatedData: Tuple that contains simulated factors (`N x K`),
    """

    # interger seed
    if isinstance(seed, int) is False:
        raise ValueError(f"seed should be an interger: received seed = {seed}")

    rng_key = random.PRNGKey(seed)
    rng_key, b_key, beta_key, s_key, var_key, obs_key, beta_key_2 = random.split(rng_key, 7)

    # dimension check
    if l_dim > p_dim:
        raise ValueError(f"l_dim should be less than p: received l_dim = {l_dim}, p = {p_dim}")
    if l_dim > p_dim / z_dim:
        raise ValueError(
            f"""l_dim is smaller than p_dim/z_dim,
            please make sure each component has {l_dim} single effects"""
        )

    if l_dim <= 0:
        raise ValueError(f"l_dim should be positive: received l_dim = {l_dim}")

    if z_dim > p_dim:
        raise ValueError(f"z_dim should be less than p: received z_dim = {z_dim}, p = {p_dim}")
    if z_dim > n_dim:
        raise ValueError(f"z_dim should be less than n: received z_dim = {z_dim}, n = {n_dim}")
    if z_dim <= 0:
        raise ValueError(f"z_dim should be positive: received z_dim = {z_dim}")

    if effect_size <= 0:
        raise ValueError(f"effect size should be positive: received effect_size = {effect_size}")

    # control data
    control_size = int(n_dim * control_fraction)

    # random W
    W = jnp.zeros(shape=(z_dim, p_dim))

    for k in range(z_dim):
        W = W.at[k, (k * l_dim) : ((k + 1) * l_dim)].set(effect_size * random.normal(b_key, shape=(l_dim,)))

    # linear function to generate Z
    # perturbation effects
    beta = jnp.zeros(shape=(g_dim, z_dim))
    # effects should be sparse
    non_zero_num = int(b_sparsity * g_dim)

    for col in range(z_dim):
        beta_key, beta_key_2 = random.split(beta_key)
        # Step 2: Randomly select `non_zero_num` indices without replacement
        indices = random.choice(beta_key, g_dim, shape=(non_zero_num,), replace=False)

        # Step 3: Generate `non_zero_num` random values from a normal distribution
        random_values = random.normal(beta_key_2, (non_zero_num,))

        # Update the selected entries in the column with random values
        beta = beta.at[indices, col].set(random_values)

    # compose new G with control cell
    G_case = create_design_matrix(s_key, n_dim, g_dim)
    G_add_zero_col = jnp.insert(G_case, 0, values=0, axis=1)
    G_control = jnp.hstack((jnp.zeros(shape=(control_size, g_dim)), jnp.ones(shape=(control_size, 1))))
    G = jnp.vstack((G_add_zero_col, G_control))
    beta_add = jnp.vstack((beta, 0.5 * random.normal(beta_key_2, shape=(1, z_dim))))
    # beta_add_zero_row = jnp.insert(beta, beta.shape[0], values=0, axis=0)

    Z = G @ beta_add + random.normal(var_key, shape=(n_dim + control_size, z_dim))
    # Latent factor model
    m = Z @ W
    X = m + random.normal(obs_key, shape=(n_dim + control_size, p_dim))

    return SimulatedData(Z, W, X, G, beta_add)



from abc import abstractmethod
from dataclasses import field

from plum import dispatch

import equinox as eqx
import jax
import jax.experimental.sparse as jsparse
import jax.lax as lax
import jax.nn as nn
import jax.numpy as jnp
import lineax as lx

from jax.scipy.special import logit
from jaxtyping import Array

from .common import ModelParams
from .sparse import SparseMatrix
from .utils import kl_discrete


def _update_sparse_beta(gdx, carry):
    ZrG, gsq_diag, params = carry

    # (t x k) (k x t)
    mean_beta_g = params.mean_beta[gdx] * params.p_hat.T[gdx]

    # add gth effect back across all K dim
    ZrG = ZrG.at[:, gdx].set(ZrG[:, gdx] + gsq_diag[gdx] * mean_beta_g)

    var_beta_g = jnp.reciprocal(params.tau_beta + gsq_diag[gdx])
    mean_beta_g = ZrG[:, gdx] * var_beta_g

    eps = 1e-8
    p_hat_g = nn.sigmoid(logit(params.p[gdx]) + 0.5 * (mean_beta_g**2) / var_beta_g)
    p_hat_g = jnp.clip(p_hat_g, eps, 1 - eps)

    # residualize based on newest estimates for downstream inf
    ZrG = ZrG.at[:, gdx].set(ZrG[:, gdx] - gsq_diag[gdx] * (mean_beta_g * p_hat_g))
    params = params._replace(
        mean_beta=params.mean_beta.at[gdx].set(mean_beta_g),
        var_beta=params.var_beta.at[gdx].set(var_beta_g),
        p_hat=params.p_hat.at[:, gdx].set(p_hat_g),
    )

    return ZrG, gsq_diag, params


@dispatch
def _get_diag(G: Array) -> Array:
    return jnp.sum(G**2, axis=0)


@dispatch
def _get_diag(G: SparseMatrix) -> Array:
    return jsparse.sparsify(jnp.sum)(G.matrix**2, axis=0).todense()  # type: ignore


@dispatch
def _wgt_sumsq(G: SparseMatrix, vector: Array) -> Array:
    tmp = G.matrix * vector
    return jsparse.sparsify(jnp.sum)(tmp**2)  # type: ignore


@dispatch
def _wgt_sumsq(G: Array, vector: Array) -> Array:
    tmp = G * vector
    return jnp.sum(tmp**2)


_multi_linear_solve = eqx.filter_vmap(lx.linear_solve, in_axes=(None, 1, None))


@dispatch
def _update_dense_beta(G: Array, params: ModelParams) -> ModelParams:
    # Create linear operator on G
    G_op = lx.MatrixLinearOperator(G)

    # Use lineax's CG solver
    solver = lx.NormalCG(rtol=1e-6, atol=1e-6)
    out = _multi_linear_solve(G_op, params.mean_z, solver)

    # Updated beta
    updated_beta = out.value.T

    return params._replace(mean_beta=updated_beta)


@dispatch
def _update_dense_beta(G: SparseMatrix, params: ModelParams) -> ModelParams:
    # Use lineax's CG solver
    solver = lx.NormalCG(rtol=1e-6, atol=1e-6)

    out = jax.vmap(lambda b: lx.linear_solve(G, b, solver), in_axes=1)(params.mean_z)

    # Updated beta
    updated_beta = out.value.T
    return params._replace(mean_beta=updated_beta)


class GuideModel(eqx.Module):
    guide_data: Array | SparseMatrix
    gsq_diag: Array = field(init=False)

    def __post_init__(self):
        self.gsq_diag = _get_diag(self.guide_data)  # type: ignore

    @property
    def shape(self):
        return self.guide_data.shape

    @abstractmethod
    def weighted_sumsq(self, params: ModelParams) -> Array: ...

    @abstractmethod
    def predict(self, params: ModelParams) -> Array: ...

    @abstractmethod
    def update(self, params: ModelParams) -> ModelParams: ...

    @staticmethod
    @abstractmethod
    def update_hyperparam(params: ModelParams) -> ModelParams: ...

    @staticmethod
    @abstractmethod
    def kl_divergence(params: ModelParams) -> Array: ...


class SparseGuideModel(GuideModel):
    def predict(self, params: ModelParams) -> Array:
        return self.guide_data @ (params.mean_beta * params.p_hat.T)

    def weighted_sumsq(self, params: ModelParams) -> Array:
        mean_bb = jnp.sum((params.mean_beta**2 + params.var_beta) * params.p_hat.T, axis=1)
        return _wgt_sumsq(self.guide_data, jnp.sqrt(mean_bb))

    def update(self, params: ModelParams) -> ModelParams:
        # compute E[Z'k]G: remove the g-th effect

        # remove predicted mean
        pred = self.predict(params)
        ZrG = (params.mean_z - pred).T @ self.guide_data

        _, g_dim = params.mean_beta.shape
        _, _, params = lax.fori_loop(0, g_dim, _update_sparse_beta, (ZrG, self.gsq_diag, params))

        return params

    @staticmethod
    def update_hyperparam(params: ModelParams) -> ModelParams:
        est_var_beta = params.mean_beta**2 + params.var_beta
        u_tau_beta = jnp.sum(params.p_hat, axis=-1) / jnp.sum(est_var_beta * params.p_hat.T, axis=0)

        return params._replace(tau_beta=u_tau_beta)

    @staticmethod
    def kl_divergence(params: ModelParams) -> Array:
        # KL for each beta
        kl_beta = 0.5 * (
            (params.mean_beta**2 + params.var_beta) * params.tau_beta
            - 1
            - jnp.log(params.var_beta)
            - jnp.log(params.tau_beta)
        )
        # sum them up, weighted by posterior prob of having an effect
        kl_beta = jnp.sum(params.p_hat.T * kl_beta)

        # KL for eta selection variables
        kl_eta = kl_discrete(params.p_hat, params.p)
        return kl_beta + kl_eta


class DenseGuideModel(GuideModel):
    def predict(self, params: ModelParams) -> Array:
        return self.guide_data @ params.mean_beta

    def weighted_sumsq(self, params: ModelParams) -> Array:
        pred_z = self.predict(params)
        return jnp.sum(pred_z**2)

    def update(self, params: ModelParams) -> ModelParams:
        return _update_dense_beta(self.guide_data, params)

    @staticmethod
    def update_hyperparam(params: ModelParams) -> ModelParams:
        return params

    @staticmethod
    def kl_divergence(params: ModelParams) -> Array:
        return jnp.asarray(0.0)



import logging


_log = logging.getLogger()

def get_logger(name, path=None):
    """get logger for perturbvi progress"""
    logger = logging.getLogger(name)
    if not logger.handlers:
        # Prevent logging from propagating to the root logger
        logger.propagate = 0
        console = logging.StreamHandler()
        logger.addHandler(console)

        # if need millisecond use : %(asctime)s.%(msecs)03d
        log_format = "[%(asctime)s - %(levelname)s] %(message)s"
        date_format = "%Y-%m-%d %H:%M:%S"
        formatter = logging.Formatter(fmt=log_format, datefmt=date_format)
        console.setFormatter(formatter)

        if path is not None:
            disk_log_stream = open("{}.log".format(path), "w")
            disk_handler = logging.StreamHandler(disk_log_stream)
            logger.addHandler(disk_handler)
            disk_handler.setFormatter(formatter)

    return logger



from __future__ import annotations

from typing import Optional

from plum import dispatch

import jax
import jax.lax as lax
import jax.numpy as jnp
import jax.numpy.linalg as jnla
import lineax as lx

from jax._src.dtypes import JAXType  # ug...
from jax.experimental import sparse
from jaxtyping import Array, ArrayLike, Float, Num


_sparse_mean = sparse.sparsify(jnp.mean)


@jax.jit
@sparse.sparsify
def _get_mean_terms(geno: Array, covar: Array) -> Array:
    m, n = covar.shape
    dtype = covar.dtype
    rcond = jnp.finfo(dtype).eps * max(n, m)
    u, s, vt = jnla.svd(covar, full_matrices=False)
    mask = s >= jnp.array(rcond, dtype=s.dtype) * s[0]

    safe_s = jnp.where(mask, s, 1).astype(covar.dtype)
    s_inv = jnp.where(mask, 1 / safe_s, 0)[:, jnp.newaxis]
    uTb = jnp.matmul(u.conj().T, geno, precision=lax.Precision.HIGHEST)

    beta = jnp.matmul(vt.conj().T, s_inv * uTb, precision=lax.Precision.HIGHEST)

    return beta


def _get_dense_var(geno: sparse.JAXSparse, dense_dtype: JAXType):
    # def _inner(_, variant):
    #     var_idx = jnp.mean(variant **2) - jnp.mean(variant) ** 2
    #     return _, var_idx
    #
    # _, var_geno = lax.scan(_inner, 0.0, geno.T)
    var_geno = _sparse_mean(geno**2, axis=0, dtype=dense_dtype) - _sparse_mean(geno, axis=0, dtype=dense_dtype) ** 2

    return var_geno.todense()


class SparseMatrix(lx.AbstractLinearOperator):
    matrix: sparse.JAXSparse

    def __init__(self, matrix: sparse.JAXSparse):
        self.matrix = matrix

    def mv(self, vector: ArrayLike):
        return sparse.sparsify(jnp.matmul)(self.matrix, vector, precision=lax.Precision.HIGHEST)  # type: ignore

    def mm(self, matrix: Num[ArrayLike, "p k"]) -> Float[Array, "n k"]:
        return jax.vmap(self.mv, (1,), 1)(matrix)

    @dispatch
    def __matmul__(self, other: lx.AbstractLinearOperator) -> lx.AbstractLinearOperator:
        return self.matrix @ other

    @dispatch
    def __matmul__(self, vector: Num[ArrayLike, " p"]) -> Float[Array, " n"]:
        return self.mv(vector)

    @dispatch
    def __matmul__(self, matrix: Num[ArrayLike, "p k"]) -> Float[Array, "p k"]:
        return self.mm(matrix)

    @dispatch
    def __rmatmul__(self, other: lx.AbstractLinearOperator) -> lx.AbstractLinearOperator:
        return other @ self.matrix

    @dispatch
    def __rmatmul__(self, vector: Num[ArrayLike, " n"]) -> Float[Array, " p"]:
        return self.T.mv(jnp.asarray(vector).T).T

    @dispatch
    def __rmatmul__(self, matrix: Num[ArrayLike, "k n"]) -> Float[Array, "k p"]:
        return self.T.mm(jnp.asarray(matrix).T).T

    def as_matrix(self) -> Float[Array, "n p"]:
        # raise ValueError("Refusing to materialise sparse matrix.")
        # Or you could do:
        return self.matrix.todense()

    def transpose(self) -> "SparseMatrix":
        return SparseMatrix(self.matrix.T)

    def in_structure(self) -> jax.ShapeDtypeStruct:
        _, in_size = self.matrix.shape
        return jax.ShapeDtypeStruct((in_size,), self.matrix.dtype)

    def out_structure(self) -> jax.ShapeDtypeStruct:
        out_size, _ = self.matrix.shape
        return jax.ShapeDtypeStruct((out_size,), self.matrix.dtype)

    @property
    def shape(self):
        n, *_ = self.out_structure().shape
        p, *_ = self.in_structure().shape

        return n, p


class CenteredSparseMatrix(lx.AbstractLinearOperator):
    data: lx.AbstractLinearOperator

    @dispatch
    def __init__(self, data: lx.AbstractLinearOperator):
        self.data = data

    @dispatch
    def __init__(
        self,
        matrix: sparse.JAXSparse,
        covar: Optional[ArrayLike] = None,
        scale: bool = False,
    ):
        n, p = matrix.shape
        geno_op = SparseMatrix(matrix)
        dtype = matrix.dtype

        if covar is None:
            covar = jnp.ones((n, 1), dtype=dtype)
            beta = _sparse_mean(matrix, axis=0, dtype=dtype).todense()
            beta = beta.reshape((1, p))
        else:
            beta = _get_mean_terms(matrix, covar)

        center_op = lx.MatrixLinearOperator(covar) @ lx.MatrixLinearOperator(beta)

        if scale:
            wgt = jnp.sqrt(_get_dense_var(matrix, dtype))
            scale_op = lx.DiagonalLinearOperator(1.0 / wgt)
            self.data = (geno_op - center_op) @ scale_op
        else:
            self.data = geno_op - center_op

    @property
    def dense_dtype(self) -> JAXType:
        return self.out_structure().dtype

    def mv(self, vector: Num[ArrayLike, " p"]) -> Float[Array, " n"]:
        return self.data.mv(vector)

    def mm(self, matrix: Num[ArrayLike, "p k"]) -> Float[Array, "n k"]:
        return jax.vmap(self.data.mv, (1,), 1)(jnp.asarray(matrix))

    @dispatch
    def __matmul__(self, other: lx.AbstractLinearOperator) -> lx.AbstractLinearOperator:
        return self.data @ other

    @dispatch
    def __matmul__(self, vector: Num[ArrayLike, " p"]) -> Float[Array, " n"]:
        return self.mv(vector)

    @dispatch
    def __matmul__(self, matrix: Num[ArrayLike, "p k"]) -> Float[Array, "p k"]:
        return self.mm(matrix)

    @dispatch
    def __rmatmul__(self, other: lx.AbstractLinearOperator) -> lx.AbstractLinearOperator:
        return other @ self.data

    @dispatch
    def __rmatmul__(self, vector: Num[ArrayLike, " n"]) -> Float[Array, " p"]:
        return self.T.mv(jnp.asarray(vector).T).T

    @dispatch
    def __rmatmul__(self, matrix: Num[ArrayLike, "k n"]) -> Float[Array, "k p"]:
        return self.T.mm(jnp.asarray(matrix).T).T

    def as_matrix(self) -> Float[Array, "n p"]:
        return self.data.as_matrix()

    def transpose(self) -> "CenteredSparseMatrix":
        return CenteredSparseMatrix(self.data.T)

    @property
    def shape(self):
        n, *_ = self.out_structure().shape
        p, *_ = self.in_structure().shape

        return n, p

    def in_structure(self) -> jax.ShapeDtypeStruct:
        return self.data.in_structure()

    def out_structure(self) -> jax.ShapeDtypeStruct:
        return self.data.out_structure()


@lx.is_symmetric.register(SparseMatrix)
@lx.is_symmetric.register(CenteredSparseMatrix)
def _(op):
    return False


@lx.is_negative_semidefinite.register(SparseMatrix)
@lx.is_negative_semidefinite.register(CenteredSparseMatrix)
def _(op):
    return False


@lx.linearise.register(SparseMatrix)
@lx.linearise.register(CenteredSparseMatrix)
def _(op):
    return op


@lx.conj.register(SparseMatrix)
def _(op):
    return SparseMatrix(op.matrix)


@lx.conj.register(CenteredSparseMatrix)
def _(op):
    return CenteredSparseMatrix(op.data)
